import base64
import json
import logging
import os.path
from datetime import datetime
from urllib.parse import urlparse
from urllib.request import urlopen

import falcon
from OpenSSL import crypto

from . import util


logger = logging.getLogger(__name__)
cert_cache = {}


class FalconAskMiddleware(object):
    """Alexa Skill Kit Middleware."""

    def __init__(self, resource, validate=True):
        self.resource = resource
        self.validate = validate

    def _get_certificate(self, cert_url):
        global cert_cache
        if cert_url in cert_cache:
            cert = cert_cache[cert_url]
            if not cert.has_expired():
                return cert
            cert_cache = {}

        # Sanity check location so we don't get some random person's cert.
        url = urlparse(cert_url)
        host = url.netloc.lower()
        path = os.path.normpath(url.path)
        if url.scheme != 'https' or \
                host not in ['s3.amazonaws.com', 's3.amazonaws.com:443'] or \
                not path.startswith('/echo.api/'):
            logger.error('Invalid certificate location: %s.', cert_url)
            return None

        resp = urlopen(cert_url)
        if resp.getcode() != 200:
            logger.error('Failed to download certificate.')
            return None

        cert = crypto.load_certificate(crypto.FILETYPE_PEM, resp.read())

        if cert.has_expired() or \
                cert.get_subject().CN != 'echo-api.amazon.com':
            logger.error('Certificate expired or invalid.')
            return None

        cert_cache[cert_url] = cert
        return cert

    def validate_request_certificate(self, req, body):
        cert_url = req.get_header('SignatureCertChainUrl', required=True)
        sig = base64.b64decode(req.get_header('Signature', required=True))

        # FU Bezos ¯\_(ツ)_/¯
        cert = self._get_certificate(cert_url)
        if not cert:
            return False

        data = req.context['raw_body']
        try:
            crypto.verify(cert, sig, data, 'sha1')
            return True
        except Exception:
            logger.exception('Invalid request signature.')
            return False

    def validate_request_timestamp(self, body, max_diff=150):
        time_str = body.get('request', {}).get('timestamp')
        if not time_str:
            logger.error('Timestamp not found: %s', body)
            return False

        req_ts = datetime.strptime(time_str, "%Y-%m-%dT%H:%M:%SZ")
        diff = (datetime.utcnow() - req_ts).total_seconds()
        if abs(diff) > max_diff:
            logger.error('Timestamp difference too high: %d sec', diff)
            return False

        return True

    def process_resource(self, req, resp, resource, params):
        if not isinstance(resource, self.resource):
            return

        allowed_methods = ['POST']
        if req.method not in allowed_methods:
            raise falcon.HTTPMethodNotAllowed(allowed_methods)

        # TODO: Validate application ID if intended for this service.

        if self.validate:
            body = req.context.get('doc') or util._get_json_body(req)
            valid_cert = self.validate_request_certificate(req, body)
            valid_ts = self.validate_request_timestamp(body)
            if not valid_cert or not valid_ts:
                logger.error('Failed to validate request.')
                raise falcon.HTTPForbidden()

        welcome = getattr(resource, 'welcome', None)
        req.context['welcome'] = welcome

        intent_maps = getattr(resource, 'intent_maps', None)
        if intent_maps is None:
            logger.error('Missing attribute "intent_maps" in resource.')
            raise falcon.HTTPInternalServerError()
        req.context['intent_maps'] = intent_maps

    def process_response(self, req, resp, resource, req_succeeded):
        """Always return valid json response.

        If resp.body is not valid json, e.g. string, do respond properly.
        """
        body = json.loads(resp.body or '{}')
        if type(body) == str:
            # Safer not to end sesion.
            response = util.respond(body, end_session=False)
            resp.body = json.dumps(response)



import json
import logging

import falcon


logger = logging.getLogger(__name__)


def _get_json_body(req):
    # req.stream corresponds to the WSGI wsgi.input environ variable,
    # and allows you to read bytes from the request body.
    #
    # See also: PEP 3333
    if req.content_length in (None, 0):
        # Nothing to do
        return None

    body = req.stream.read()
    if not body:
        raise falcon.HTTPBadRequest(
            'Empty request body',
            'A valid JSON document is required.'
        )

    try:
        data = json.loads(body.decode('utf-8'))
        req.context['doc'] = data
        req.context['raw_body'] = body
        return data
    except (ValueError, UnicodeDecodeError):
        raise falcon.HTTPError(
           falcon.HTTP_753,
           'Malformed JSON',
           'Could not decode the request body. The '
           'JSON was incorrect or not encoded as '
           'UTF-8.'
        )


def get_req_type(body):
    return body.get('request', {}).get('type')


def get_slots(body):
    if get_req_type(body) != 'IntentRequest':
        return {}

    slots = body['request']['intent'].get('slots', {})
    return {item['name']: item.get('value') for _, item in slots.items()}


def dispatch_request(req):
    # Check if req.stream is consumed, otherwise get POST data.
    body = req.context.get('doc') or _get_json_body(req)
    logger.debug('Got body: %s' % body)

    req_type = get_req_type(body)
    intent_maps = req.context['intent_maps']

    if req_type == 'LaunchRequest':
        return req.context['welcome']
    elif req_type == 'IntentRequest':
        intent = body['request']['intent']['name']
        try:
            intent_fn = intent_maps[intent]
        except KeyError:
            logger.exception(
                'No intent function mapping found for intent: %s' % intent
            )
            try:
                intent_fn = intent_maps['UnknownIntent']
            except KeyError:
                logger.exception('UnknownIntent mapping also not found.')
                raise falcon.HTTPInternalServerError()
        return intent_fn(body)
    elif req_type == 'SessionEndedRequest':
        logger.info('SessionEndedRequest received. Session ended.')
        return None

    logger.error('Invalid request type: %s', req_type)
    raise falcon.HTTPBadRequest()


def respond(text=None, ssml=None, attributes=None, reprompt_text=None,
            reprompt_ssml=None, end_session=True, directives=None):
    obj = {
        'version': '1.0',
        'response': {
            'shouldEndSession': end_session
        },
        'sessionAttributes': attributes or {}
    }

    if text:
        obj['response']['outputSpeech'] = {'type': 'PlainText', 'text': text}
    elif ssml:
        obj['response']['outputSpeech'] = {'type': 'SSML', 'ssml': ssml}

    reprompt_output = None
    if reprompt_text:
        reprompt_output = {'type': 'PlainText', 'text': reprompt_text}
    elif reprompt_ssml:
        reprompt_output = {'type': 'SSML', 'ssml': reprompt_ssml}

    if reprompt_output:
        obj['response']['reprompt'] = {'outputSpeech': reprompt_output}

    obj['response']['directives'] = directives or []

    return obj



from setuptools import setup


with open('README.md', 'r', encoding='utf-8') as f:
    readme = f.read()

setup(
    name='falcon_ask',
    version='1.1.1',
    description='Toolkit for writing Amazon Alexa skills in your Falcon app',
    long_description=readme,
    long_description_content_type='text/markdown',
    author='Philamer Sune',
    author_email='pvsune@hotmail.com',
    url='https://github.com/pvsune/falcon_ask',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: ISC License (ISCL)',
        'Programming Language :: Python :: 3.5',
        'Topic :: Home Automation',
        'Topic :: Internet :: WWW/HTTP :: WSGI :: Application',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
    ],
    packages=['falcon_ask'],
    license='ISC',
    tests_require=['pytest'],
    install_requires=[
        'pyOpenssl==17.3.0'
    ]
)
